"""Tools and base classes for implementing full IOTile devices in software

An IOTile device is defined as something you can connect to with
HardwareManager and interact with via its rpc, streaming, tracing, script, and
debug interfaces.

Many IOTile devices are physical, with an embedded microcontroller running
code that responds to requests on one of the above interfaces.

However, since the connection between HardwareManager and the device is
entirely mediated by an `class`:AbstractDeviceAdapter`, there is no need for
the device to be a physical thing.  If you have a device adapter that routes
requests from HardwareManager to a software program such as a TCP or websockets
server then there is no need for physical hardware.

The classes and functions in this subpackage provide base implementations of
what is needed to create a software-only, "virtual", iotile device.

.. important::

    Just because virtual devices run in python, does not mean that you cannot
    interact with them from outside your computer.  If you attach a
    DeviceServer class to a virtual device then you can access that device via
    whatever mechanism the DeviceServer supports, which could be, for example,
    bluetooth, usb or wifi.

Put another way, this subpackage provides the resources to be able to
implement the *server* side of an IOTile device, referring to
``HardwareManager`` and most of CoreTools as the *client* side.

There are other server implementations such as one written in embedded C and
not included with CoreTools that serve an equivalent purpose.

The bottom line is that this package is designed to let you build iotile
devices in software without needing IOTile-specific hardware.


Developer Reference
===================

The base entrypoint for creating virtual devices is the class
:class:`BaseVirtualDevice`. It defines the bare minimum of what makes the
device iotile-compatible, including:

- supporting iotile interfaces that can be opened and closed
- having a way for clients to call an rpc
- having support for the device asynchronously push data to the user via its
  streaming and tracing interfaces.
- having a way for the device to receive a script from the client

.. note::

    There is no reference to "tiles" in the above description of a virtual device.
    Tiles, defined as subsystems of an iotile device with a unique address, are
    not a fundamental concept.  They just happen to be a useful way of organizing
    complex iotile devices into reuable modules.

All virtual devices that you create must inherit from :class:`BaseVirtualDevice`.
At a minimum you need to override a single function:

- :meth:`BaseVirtualDevice.call_rpc`: This is the method that will be called
  when a client wants to execute an RPC on your device.  It must execute the
  RPC and return the result or raise an exception.  If the RPC is implemented
  as a coroutine, it may return a coroutine or awaitable object containing the
  result / exception.

.. warning::

    Most developers will not want to inherit directly from
    :class:`BaseVirtualDevice`, because it doesn't provide convenience
    functionality that you likely want to use.  For that reason there are two
    premade subclasses: :class:`SimpleVirtualDevice` and
    :class:`StandardVirtualDevice`.

    ``SimpleVirtualDevice`` does not have support for tiles, it just allows
    you register RPCs with a fixed address and rpc_id on the device itself.
    It is useful for very simple things where you want a device that
    implements very few RPCs.

    ``StandardVirtualDevice`` requires that you create :class:`VirtualTile`
    clases that contain all of your RPC definitions and it allows you to load
    those tiles into a given address in the virtual device and then RPCs are
    delegated to the tiles.

    Most developers will want to inherit from :class:`StandardVirtualDevice`.


Virtual Device Lifecycle
------------------------

All virtual devices are run inside of a :class:`VirtualDeviceAdapter`, even
when they are served externally using a DeviceServer.  The DeviceServer
interacts with the VirtualDeviceAdapter, which in turn interacts with the
virtual device.

The lifecycle of a virtual device, then is entirely defined by the
VirtualDeviceAdapter that owns it.  In general the flow is as follows:

- The virtual device instance is created.  At this point it is not running,
  cannot respond to RPCs and no background work should be started.

- ``start(channel)`` is called on the virtual device.  This is equivalent to the
  virtual device "turning on".  It should start any background work that
  it needs to do and prepare to receive RPCs.  The ``channel`` paramater passed
  from ``VirtualDeviceAdapter`` to virtual device is an instance of
  :class:`VirtualAdapterAsyncChannel`, which has public methods that the device
  can call at any time to asynchronously push data to the user through its
  streaming and tracing interfaces.

- ``stop()`` is called when the virtual device should be stopped.  It is
  equivalent to powering down the virtual device.  All background work should
  be synchronously stopped and no more RPCs will be dispatched to the device.
  Any previously connected clients will have already been disconnected.

"""

from .virtualtile import VirtualTile
from .virtualtile_base import BaseVirtualTile
from .virtualdevice_base import BaseVirtualDevice, AbstractAsyncDeviceChannel
from .virtualdevice_standard import StandardVirtualDevice
from .virtualdevice_simple import SimpleVirtualDevice

from .common_types import (rpc, tile_rpc, unpack_rpc_payload, pack_rpc_payload,
                           pack_rpc_response, unpack_rpc_response, RPCDispatcher,
                           RPCDeclaration)

__all__ = ['BaseVirtualTile', 'VirtualTile', 'BaseVirtualDevice',
           'AbstractAsyncDeviceChannel', 'StandardVirtualDevice',
           'RPCDeclaration', 'SimpleVirtualDevice', 'tile_rpc', 'rpc',
           'unpack_rpc_payload', 'pack_rpc_payload', 'pack_rpc_response',
           'unpack_rpc_response']



"""Mixin for keeping track of per-connection data."""

from ...exceptions import DeviceAdapterError

#pylint:disable=too-few-public-methods;This is a mixin class
class PerConnectionDataMixin:
    """A mixin class for DeviceAdapters to track per-connection data.

    This class helps adapters comply with the requirement to be able to track
    per-connection information for internal usage as well as providing a
    reference implementation of :meth:`_get_conn_id`, which is needed to map
    connection_strings to connection_ids for the purposes of sending
    notifications attached to a connection.

    It is very simple and requires that your adapter's ``connect`` and ``disconnect``
    implementations call :meth:`_setup_connection` and :meth:`_teardown_connection`
    respectively in order to manage the connection metadata correctly.

    If you receive an unexpected disconnection event from a device, you must
    also call :meth:`_teardown_connection` as well to make sure the connection
    state is properly recorded.

    In any method that requires a valid open connection, you can just call
    :meth:`_ensure_connection` at the start of your implementation to fail correctly
    if the connection is no longer valid.
    """

    def __init__(self):
        self._connections = {}
        self._reverse_connections = {}

    def _get_property(self, conn_id, name):
        info = self._connections.get(conn_id)
        if info is None:
            raise DeviceAdapterError(conn_id, 'none', 'connection id does not exist')

        return info.get(name)

    def _track_property(self, conn_id, name, value):
        info = self._connections.get(conn_id)
        if info is None:
            raise DeviceAdapterError(conn_id, 'none', 'connection id does not exist')

        info[name] = value

    def _ensure_connection(self, conn_id, is_connected):
        if (conn_id in self._connections) != is_connected:
            raise DeviceAdapterError(conn_id, 'none', 'connection in wrong state')

    def _setup_connection(self, conn_id, conn_string):
        if conn_id in self._connections:
            raise DeviceAdapterError(conn_id, 'none', 'connection id setup twice')

        self._connections[conn_id] = dict(connection_string=conn_string)
        self._reverse_connections[conn_string] = conn_id

    def _teardown_connection(self, conn_id, force=False):
        if conn_id not in self._connections:
            if force:
                return

            raise DeviceAdapterError(conn_id, 'none', 'connection id torndown without being setup')

        conn_string = self._connections[conn_id]['connection_string']
        del self._connections[conn_id]
        del self._reverse_connections[conn_string]

    def _get_conn_id(self, conn_string):
        return self._reverse_connections.get(conn_string, None)



import struct
from .rootkey_auth_provider import RootKeyAuthProvider

from iotile.core.exceptions import NotFoundError

class NullAuthProvider(RootKeyAuthProvider):
    """Auth provider that use null root key, uuid extended to 32 bytes with zeroes"""

    def __init__(self, args=None):
        if args is None:
            args = {}

        args['supported_keys'] = [self.NullKey]

        super().__init__(args)

    def get_root_key(self, key_type, device_id):
        self.verify_key(key_type)

        return bytearray(16)



"""An adapter class that takes a DeviceAdapter and produces a CMDStream compatible interface"""

from copy import deepcopy
import queue
from time import monotonic, sleep
from datetime import datetime
import binascii
import logging
import threading
from iotile.core.exceptions import HardwareError, ArgumentError
from iotile.core.utilities import SharedLoop
from ..exceptions import VALID_RPC_EXCEPTIONS
from ..virtual import unpack_rpc_response, pack_rpc_response
from .adapter import AbstractDeviceAdapter, AsynchronousModernWrapper, DeviceAdapter


class _RecordedRPC:
    """Internal helper class for saving recorded RPCs to csv files."""

    def __init__(self, connection, address, rpc_id, call):
        if isinstance(connection, bytes):
            connection = connection.decode('utf-8')

        self.connection = connection
        self.address = address
        self.rpc_id = rpc_id
        self.call = binascii.hexlify(call).decode('utf-8')

        self.status = -1
        self.start_stamp = None

        self.response = ""
        self.error = ""

        self.runtime = 0
        self._start_time = 0

    def start(self):
        """Mark the beginning of a recorded RPC."""

        self._start_time = monotonic()
        self.start_stamp = datetime.utcnow()

    def finish(self, status, response):
        """Mark the end of a recorded RPC."""

        self.response = binascii.hexlify(response).decode('utf-8')
        self.status = status
        self.runtime = monotonic() - self._start_time

    def serialize(self):
        """Convert this recorded RPC into a string."""

        return "{},{: <26},{:2d},{:#06x},{:#04x},{:5.0f},{: <40},{: <40},{}".\
            format(self.connection, self.start_stamp.isoformat(), self.address, self.rpc_id,
                   self.status, self.runtime * 1000, self.call, self.response, self.error)


class AdapterStream:
    """A wrapper that provides a synchronous interface on top of AbstractDeviceAdapter.

    DeviceAdapters have a more generic interface that is not restricted to
    exclusive use in an online fashion by a single user at a time.  This class
    implements a simpler interface on top of an AbstractDeviceAdapter that is
    easier to use from programs that only talk to a single device at a time or
    only have a single user interacting at a time.

    AdapterStream is the internal implementation that powers :class:`HardwareManager`
    and has a very similar interface.  In constrast to a DeviceAdapter which is
    command and even driven and stores very little data for a given device, AdapterStream
    contains queues that buffer reports, traces and broadcasts received from devices
    in order to make it more convenient to process them.  It also contains a generic
    RPC logging mechanism to produce a record of what RPCs were sent to a device.

    Args:
        adapter (AbstractDeviceAdapter): the DeviceAdapter that we should use
        record (string): The path to a file that we should use to record any RPCs
            sent for tracing purposes.
    """

    def __init__(self, adapter, record=None, loop=SharedLoop):
        self._scanned_devices = {}
        self._scan_lock = threading.Lock()
        self._reports = None
        self._broadcast_reports = None
        self._traces = None

        self._loop = loop
        self._record = record
        if self._record is not None:
            self._recording = []

        self._on_progress = None

        self._logger = logging.getLogger(__name__)

        self.connection_interrupted = False
        self.connected = False
        self.connection_string = None

        if isinstance(adapter, DeviceAdapter):
            adapter = AsynchronousModernWrapper(adapter)
        elif not isinstance(adapter, AbstractDeviceAdapter):
            raise ArgumentError("Unknown class passed as DeviceAdapter: %s" % str(adapter), adapter=adapter)

        self.adapter = adapter
        self._loop.run_coroutine(adapter.start())

        self.adapter.register_monitor([None], ['report', 'trace', 'broadcast', 'disconnection', 'device_seen', 'progress'],
                                      self._on_notification)

        self._start_time = monotonic()

    def scan(self, wait=None):
        """Return the devices that have been found for this device adapter.

        If the adapter indicates that we need to explicitly tell it to probe for devices, probe now.
        By default we return the list of seen devices immediately, however there are two cases where
        we will sleep here for a fixed period of time to let devices show up in our result list:

        - If we are probing then we wait for 'minimum_scan_time'
        - If we are told an explicit wait time that overrides everything and we wait that long
        """

        min_scan = self.adapter.get_config('minimum_scan_time', 0.0)
        probe_required = self.adapter.get_config('probe_required', False)

        # Figure out how long and if we need to wait before returning our scan results
        wait_time = None
        elapsed = monotonic() - self._start_time
        if elapsed < min_scan:
            wait_time = min_scan - elapsed

        # If we need to probe for devices rather than letting them just bubble up, start the probe
        # and then use our min_scan_time to wait for them to arrive via the normal _on_scan event
        if probe_required:
            self._loop.run_coroutine(self.adapter.probe())
            wait_time = min_scan

        # If an explicit wait is specified that overrides everything else
        if wait is not None:
            wait_time = wait

        if wait_time is not None:
            sleep(wait_time)

        to_remove = set()

        now = monotonic()

        with self._scan_lock:
            for name, value in self._scanned_devices.items():
                if value['expiration_time'] < now:
                    to_remove.add(name)

            for name in to_remove:
                del self._scanned_devices[name]

            devices = sorted(self._scanned_devices.values(), key=lambda x: x['uuid'])

        return devices

    def connect(self, uuid_value, wait=None):
        """Connect to a specific device by its uuid

        Attempt to connect to a device that we have previously scanned using its UUID.
        If wait is not None, then it is used in the same was a scan(wait) to override
        default wait times with an explicit value.

        Args:
            uuid_value (int): The unique id of the device that we would like to connect to.
            wait (float): Optional amount of time to force the device adapter to wait before
                attempting to connect.
        """

        if self.connected:
            raise HardwareError("Cannot connect when we are already connected")

        if uuid_value not in self._scanned_devices:
            self.scan(wait=wait)

        with self._scan_lock:
            if uuid_value not in self._scanned_devices:
                raise HardwareError("Could not find device to connect to by UUID", uuid=uuid_value)

            connstring = self._scanned_devices[uuid_value]['connection_string']

        self.connect_direct(connstring)

    def connect_direct(self, connection_string, no_rpc=False, force=False):
        """Directly connect to a device using its stream specific connection string.

        Normally, all connections to a device include opening the RPC
        interface to send RPCs.  However, there are certain, very specific,
        circumstances when you would not want to or be able to open the RPC
        interface (such as when you are using the debug interface on a bare
        MCU that has not been programmed yet).  In those cases you can pass
        no_rpc=True to not attempt to open the RPC interface.  If you do not
        open the RPC interface at connection time, there is no public
        interface to open it later, so you must disconnect and reconnect to
        the device in order to open the interface.

        Args:
            connection_string (str): The connection string that identifies the desired device.
            no_rpc (bool): Do not open the RPC interface on the device (default=False).
            force (bool): Whether to force another connection even if we think we are currently
                connected.  This is for internal use and not designed to be set externally.
        """

        if not force and self.connected:
            raise HardwareError("Cannot connect when we are already connected to '%s'" % self.connection_string)

        self._loop.run_coroutine(self.adapter.connect(0, connection_string))

        try:
            if no_rpc:
                self._logger.info("Not opening RPC interface on device %s", self.connection_string)
            else:
                self._loop.run_coroutine(self.adapter.open_interface(0, 'rpc'))
        except HardwareError as exc:
            self._logger.exception("Error opening RPC interface on device %s", connection_string)
            self._loop.run_coroutine(self.adapter.disconnect(0))
            raise exc
        except Exception as exc:
            self._logger.exception("Error opening RPC interface on device %s", connection_string)
            self._loop.run_coroutine(self.adapter.disconnect(0))
            raise HardwareError("Could not open RPC interface on device due to an exception: %s" % str(exc)) from exc

        self.connected = True
        self.connection_string = connection_string
        self.connection_interrupted = False

    def disconnect(self):
        """Disconnect from the device that we are currently connected to."""

        if not self.connected:
            raise HardwareError("Cannot disconnect when we are not connected")

        # Close the streaming and tracing interfaces when we disconnect
        self._reports = None
        self._traces = None

        if not self.connection_interrupted:
            self._loop.run_coroutine(self.adapter.disconnect(0))
        self.connected = False
        self.connection_interrupted = False
        self.connection_string = None

    def _try_reconnect(self):
        """Try to recover an interrupted connection."""

        try:
            if self.connection_interrupted:
                self.connect_direct(self.connection_string, force=True)
                self.connection_interrupted = False
                self.connected = True

                # Reenable streaming interface if that was open before as well
                if self._reports is not None:
                    self._loop.run_coroutine(self.adapter.open_interface(0, 'streaming'))

                # Reenable tracing interface if that was open before as well
                if self._traces is not None:
                    self._loop.run_coroutine(self.adapter.open_interface(0, 'tracing'))
        except HardwareError as exc:
            self._logger.exception("Error reconnecting to device after an unexpected disconnect")
            raise HardwareError("Device disconnected unexpectedly and we could not reconnect", reconnect_error=exc) from exc

    def send_rpc(self, address, rpc_id, call_payload, timeout=3.0):
        """Send an rpc to our connected device.

        The device must already be connected and the rpc interface open.  This
        method will synchronously send an RPC and wait for the response.  Any
        RPC errors will be raised as exceptions and if there were no errors, the
        RPC's response payload will be returned as a binary bytearray.

        See :meth:`AbstractDeviceAdapter.send_rpc` for documentation of the possible
        exceptions that can be raised here.

        Args:
            address (int): The tile address containing the RPC
            rpc_id (int): The ID of the RPC that we wish to call.
            call_payload (bytes): The payload containing encoded arguments for the
                RPC.
            timeout (float): The maximum number of seconds to wait for the RPC to
                finish.  Defaults to 3s.

        Returns:
            bytearray: The RPC's response payload.
        """

        if not self.connected:
            raise HardwareError("Cannot send an RPC if we are not in a connected state")

        if timeout is None:
            timeout = 3.0

        status = -1
        payload = b''
        recording = None

        if self.connection_interrupted:
            self._try_reconnect()

        if self._record is not None:
            recording = _RecordedRPC(self.connection_string, address, rpc_id, call_payload)
            recording.start()

        try:
            payload = self._loop.run_coroutine(self.adapter.send_rpc(0, address, rpc_id, call_payload, timeout))
            status, payload = pack_rpc_response(payload, None)
        except VALID_RPC_EXCEPTIONS as exc:
            status, payload = pack_rpc_response(payload, exc)

        if self._record is not None:
            recording.finish(status, payload)
            self._recording.append(recording)

        if self.connection_interrupted:
            self._try_reconnect()

        return unpack_rpc_response(status, payload, rpc_id, address)

    def send_highspeed(self, data, progress_callback):
        """Send a script to a device at highspeed, reporting progress.

        This method takes a binary blob and downloads it to the device as fast
        as possible, calling the passed progress_callback periodically with
        updates on how far it has gotten.

        Args:
            data (bytes): The binary blob that should be sent to the device at highspeed.
            progress_callback (callable): A function that will be called periodically to
                report progress.  The signature must be callback(done_count, total_count)
                where done_count and total_count will be passed as integers.
        """

        if not self.connected:
            raise HardwareError("Cannot send a script if we are not in a connected state")

        if isinstance(data, str) and not isinstance(data, bytes):
            raise ArgumentError("You must send bytes or bytearray to _send_highspeed", type=type(data))

        if not isinstance(data, bytes):
            data = bytes(data)

        try:
            self._on_progress = progress_callback
            self._loop.run_coroutine(self.adapter.send_script(0, data))
        finally:
            self._on_progress = None

    def enable_streaming(self):
        """Open the streaming interface and accumute reports in a queue.

        This method is safe to call multiple times in a single device
        connection. There is no way to check if the streaming interface is
        opened or to close it once it is opened (apart from disconnecting from
        the device).

        The first time this method is called, it will open the streaming
        interface and return a queue that will be filled asynchronously with
        reports as they are received.  Subsequent calls will just empty the
        queue and return the same queue without interacting with the device at
        all.

        Returns:
            queue.Queue: A queue that will be filled with reports from the device.
        """

        if not self.connected:
            raise HardwareError("Cannot enable streaming if we are not in a connected state")

        if self._reports is not None:
            _clear_queue(self._reports)
            return self._reports

        self._reports = queue.Queue()
        self._loop.run_coroutine(self.adapter.open_interface(0, 'streaming'))

        return self._reports

    def enable_tracing(self):
        """Open the tracing interface and accumulate traces in a queue.

        This method is safe to call multiple times in a single device
        connection. There is no way to check if the tracing interface is
        opened or to close it once it is opened (apart from disconnecting from
        the device).

        The first time this method is called, it will open the tracing
        interface and return a queue that will be filled asynchronously with
        reports as they are received.  Subsequent calls will just empty the
        queue and return the same queue without interacting with the device at
        all.

        Returns:
            queue.Queue: A queue that will be filled with trace data from the device.

            The trace data will be in disjoint bytes objects in the queue
        """

        if not self.connected:
            raise HardwareError("Cannot enable tracing if we are not in a connected state")

        if self._traces is not None:
            _clear_queue(self._traces)
            return self._traces

        self._traces = queue.Queue()
        self._loop.run_coroutine(self.adapter.open_interface(0, 'tracing'))

        return self._traces

    def enable_broadcasting(self):
        """Begin accumulating broadcast reports received from all devices.

        This method will allocate a queue to receive broadcast reports that
        will be filled asynchronously as broadcast reports are received.

        Returns:
            queue.Queue: A queue that will be filled with braodcast reports.
        """

        if self._broadcast_reports is not None:
            _clear_queue(self._broadcast_reports)
            return self._broadcast_reports

        self._broadcast_reports = queue.Queue()
        return self._broadcast_reports

    def enable_debug(self):
        """Open the debug interface on the connected device."""

        if not self.connected:
            raise HardwareError("Cannot enable debug if we are not in a connected state")

        self._loop.run_coroutine(self.adapter.open_interface(0, 'debug'))

    def debug_command(self, cmd, args=None, progress_callback=None):
        """Send a debug command to the connected device.

        This generic method will send a named debug command with the given
        arguments to the connected device.  Debug commands are typically used
        for things like forcible reflashing of firmware or other, debug-style,
        operations.  Not all transport protocols support debug commands and
        the supported operations vary depeneding on the transport protocol.

        Args:
            cmd (str): The name of the debug command to send.
            args (dict): Any arguments required by the given debug command
            progress_callback (callable): A function that will be called periodically to
                report progress.  The signature must be callback(done_count, total_count)
                where done_count and total_count will be passed as integers.

        Returns:
            object: The return value of the debug command, if there is one.
        """

        if args is None:
            args = {}

        try:
            self._on_progress = progress_callback
            return self._loop.run_coroutine(self.adapter.debug(0, cmd, args))
        finally:
            self._on_progress = None

    def close(self):
        """Close this adapter stream.

        This method may only be called once in the lifetime of an
        AdapterStream and it will shutdown the underlying device adapter,
        disconnect all devices and stop all background activity.

        If this stream is configured to save a record of all RPCs, the RPCs
        will be logged to a file at this point.
        """

        try:
            self._loop.run_coroutine(self.adapter.stop())
        finally:
            self._save_recording()

    def _on_notification(self, conn_string, _conn_id, name, event):
        if name not in ('device_seen', 'broadcast') and conn_string != self.connection_string:
            return

        if name == 'report':
            self._on_report(event)
        elif name == 'broadcast':
            self._on_broadcast(event)
        elif name == 'trace':
            self._on_trace(event)
        elif name == 'device_seen':
            self._on_scan(event)
        elif name == 'disconnection':
            self._on_disconnect()
        elif name == 'progress' and self._on_progress is not None:
            self._on_progress(event.get('finished'), event.get('total'))

    def _on_broadcast(self, report):
        if self._broadcast_reports is None:
            return

        self._broadcast_reports.put(report)

    def _on_report(self, report):
        if self._reports is None:
            return

        self._reports.put(report)

    def _on_trace(self, tracing_data):
        if self._traces is None:
            return

        self._traces.put(tracing_data)

    def _on_scan(self, info):
        """Callback called when a new device is discovered on this CMDStream

        Args:
            info (dict): Information about the scanned device
        """

        device_id = info['uuid']
        expiration_time = info.get('validity_period', 60)
        infocopy = deepcopy(info)

        infocopy['expiration_time'] = monotonic() + expiration_time

        with self._scan_lock:
            self._scanned_devices[device_id] = infocopy

    def _on_disconnect(self):
        """Callback when a device is disconnected unexpectedly.

        Args:
            adapter_id (int): An ID for the adapter that was connected to the device
            connection_id (int): An ID for the connection that has become disconnected
        """

        self._logger.info("Connection to device %s was interrupted", self.connection_string)
        self.connection_interrupted = True

    def _save_recording(self):
        if not self._record:
            return

        with open(self._record, "w", encoding="utf-8") as outfile:
            outfile.write("# IOTile RPC Recording\n")
            outfile.write("# Format: 1.0\n\n")
            outfile.write("Connection,Timestamp [utc isoformat],Address,RPC ID,"
                          "Duration [ms],Status,Call,Response,Error\n")

            for recording in self._recording:
                outfile.write(recording.serialize())
                outfile.write('\n')

        self._record = None


def _clear_queue(to_clear):
    """Clear all items from a queue safely."""

    while not to_clear.empty():
        try:
            to_clear.get(False)
            to_clear.task_done()
        except queue.Empty:
            continue



import hashlib
from iotile.core.exceptions import ArgumentError

def calculate_sha(sha_type, data):
    """Uses sha to calculate a checksum"""

    if sha_type == 256:
        shasum = hashlib.sha256()
        shasum.update(data)

        return shasum.hexdigest()

    raise ArgumentError("Unknown/Unimplemented sha algorithm")


import crcmod
from iotile.core.exceptions import ArgumentError

def calculate_crc(crc_type, data):
    """Uses crc to calculate a checksum"""

    if crc_type == 0x104C11DB7:
        crc_func = crcmod.mkCrcFun(crc_type, initCrc=0xFFFFFFFF,
                                             rev=False, xorOut=0)
        checksum = crc_func(data) & 0xFFFFFFFF

        return hex(checksum)

    raise ArgumentError("Unknown/Unimplemented crc algorithm")



"""Abstract interface for a class that serves access to IOTile devices.

This class can be thought of as a generic server that could be accessed
by an :class:`AbstractDeviceAdapter`.  This base class does not specify
anything about how the actual server works since that will vary widely.

The goal of declaring this base class is to standardize how such servers
are configured and started so that a generic "server container" such as
``iotile-gateway`` is capable of running any number of servers regardless
of their implementations.

It is, by design, entirely unspecified how a given device server chooses
to "make available" devices.  The only agreement needs to be between
a device server and the corresponding device adapter.

As a concrete example, consider the BLED112 DeviceAdapter.  This adapter acts
as a BLE central device, finding iotile devices by scanning for bluetooth low
energy (BLE) perihperal advertisements within range.  It sends rpcs by writing
to the iotile device's GATT server on specific characteristics.

A device server for BLED112 then would be a BLE peripheral that implements the
same GATT server as a phyiscal IOTile device and dispatches RPCs to its
internal adapter as they are received via writes to the agreed-upon
characterstics.

Given the highly normalized nature of IOTile interactions, they are easy to
proxy between communications modalities and form bridges connecting, e.g.
bluetooth and websockets or TCP and usb.
"""

import abc
from iotile.core.utilities import SharedLoop

class AbstractDeviceServer(abc.ABC):
    """A generic class that serves access to an IOTile compliant device.

    This class is the server side of a :class:`AbstractDeviceAdapter`.
    It is expected that there could be an AbstractDeviceServer for
    every AbstractDeviceAdapter implementation.

    In an analogy to HTTP, the DeviceAdapter is the HTTP client like
    Chrome or curl.  The DeviceServer is the HTTP server like nginx
    or apache.

    Note that there is very little specified here in terms of required
    interface except that arguments must be passed as an argument dictionary
    and the device server must take a single adapter argument that will
    be an AbstractDeviceAdapter.  The device server should use that device
    adapter to find, connect to and perform operations on devices.

    .. important:

        It is not the role of AbstractDeviceServer to ``start()`` or ``stop()``
        its device adapter.  It should assume that the adapter is already
        started by the time :meth:`start` is called and will be stopped
        after :meth:`stop` is called.

        This is important because there may well be multiple device servers
        attached to the same underlying device adapter.

    Args:
        adapter (AbstractDeviceAdapter): The device adapter that this
            device server should use to find and connect to devices.
        args (dict): Any arguments to the device adapter must be passed
            in as a dictionary so that they can be passed in a generic
            fashion.
        loop (BackgroundEventLoop): The event loop that the device server
            should use for blocking coroutine operations.  If not specified,
            this defaults to the global SharedLoop.
    """

    @abc.abstractmethod
    def __init__(self, adapter, args=None, *, loop=SharedLoop):
        """Required constructor signature."""

    @abc.abstractmethod
    async def start(self):
        """Start serving access to devices.

        This method must not return until the devices are accessible over the
        implementations chosen server protocol.  It must be possible to
        immediately attach an AbstractDevice adapter to this server and use it
        without any race conditions.

        For example, if the server is configured to use a TCP port to serve
        devices, the TCP port must be bound and listening for connections when
        this coroutine returns.
        """

    @abc.abstractmethod
    async def stop(self):
        """Stop serving access to devices.

        Subclass **must not** return until the devices are no longer accessible
        over the implementation's chosen server protocol.  There should be no
        race condition where a client is still able to invoke a function on
        this server after stop has returned.  This is important because server
        container running this AbstractDeviceServer may no longer be in a
        state to respond allow for client connections safely.

        Subclasses **must** cleanly disconnect from any devices accessed via
        ``adapter`` before stopping.  It is the job of the ``AbstractDeviceServer``
        to release all of its internal resources.  It is not the job of ``adapter``
        to somehow know that a particular DeviceServer has stopped and free the
        resources associated with that server.

        Subclasses **should** cleanly disconnect any clients when stopped if
        possible, rather than just breaking all sockets, for example.
        """



"""DeviceAdapter that connects to virtual devices that are run inside the same process.

This adapter is very useful for integration testing other systems that depends
on device interaction since you can create whatever kind of situation you want
using a virtual device and to anyone outside of HardwareManager, there is no
way to tell they are not talking to a physical iotile device.

This adapter also forms the basis for transparent proxying of device access.
Using a combination of a DeviceServer class attached to a VirtaulDevice, you
can project any virtual device out into the world as a connectable iotile
device over any protocol supported by an installed device adapter.  This could
let you, for example, serve a virtual device from your computer over bluetooth
low energy that lets users control your computer from their mobile phone.
"""

import json
import logging
import inspect
from iotile.core.exceptions import ArgumentError
from iotile.core.dev import ComponentRegistry
from iotile.core.hw.reports import BroadcastReport
from iotile.core.hw.exceptions import DevicePushError
from iotile.core.utilities import SharedLoop
from ..exceptions import DeviceAdapterError, VALID_RPC_EXCEPTIONS
from ..virtual import BaseVirtualDevice, AbstractAsyncDeviceChannel
from .adapter import StandardDeviceAdapter


class VirtualAdapterAsyncChannel(AbstractAsyncDeviceChannel):
    """A channel for tracing and streaming data asynchronously from virtual devices"""

    def __init__(self, adapter, iotile_id):
        self.adapter = adapter
        self.iotile_id = iotile_id
        self.conn_string = str(iotile_id)

    async def stream(self, report):
        """Queue data for streaming

        Args:
            report (IOTileReport): A report object to stream to a client
        """

        conn_id = self._find_connection(self.conn_string)

        if isinstance(report, BroadcastReport):
            await self.adapter.notify_event(self.conn_string, 'broadcast', report)
        elif conn_id is not None:
            await self.adapter.notify_event(self.conn_string, 'report', report)
        else:
            raise DevicePushError("Cannot push report because no client is connected")

    async def trace(self, data):
        """Queue data for tracing

        Args:
            data (bytes): Unstructured data to trace to any
                connected client.
        """

        conn_id = self._find_connection(self.conn_string)

        if conn_id is not None:
            await self.adapter.notify_event(self.conn_string, 'trace', data)
        else:
            raise DevicePushError("Cannot push tracing data because no client is connected")

    async def disconnect(self):
        """Forcibly disconnect a connected client."""

        raise DevicePushError("Disconnection is not yet supported")

    def _find_connection(self, conn_string):
        """Find the connection corresponding to an iotile_id."""

        return self.adapter._get_conn_id(conn_string)


class VirtualDeviceAdapter(StandardDeviceAdapter):
    """Device adapter that gives access to one or more virtual devices

    The adapter is created and serves access to the virtual_devices that are
    found by name in the entry_point group iotile.virtual_device.

    There are two ways to passing configuration dictionaries down into a
    VirtualDevice attached to this VirtualDeviceAdapter.  Both are performed
    by encoding the port string with {{device_name}}@{{config_info}}.  You
    can either pass a path to a json file in {{config_info}} which is loaded
    as a dictionary or you can prepend a '#' to the front of config_info and
    pass a base64 encoded json string that is decoded into a dict.  Both are
    identically passed to the underlying __init__ function of the VirtualDevice

    If you choose to use a json file, the format must be::

        {
            "device":
            {
                ARGS GO HERE
            }
        }

    The required device key is to allow for this file to host information about
    multiple other kinds of iotile entities without collisions.

    If you are programmatically specifying the dictionary then it should just
    be a normal json with no parent "device" key such as:

        {
            ARGS GO HERE
        }

    Args:
        port (string): A port description that should be in the form of
            device_name1@<optional_config_json1;device_name2@optional_config_json2
        devices (list of BaseVirtualDevice): Optional list of precreated virtual
            devices that you would like to add to this virtual adapter.  It can sometimes
            be easier to pass precreated devices rather than needing to specify how
            they should be created.
    """

    # Make devices expire after a long time only
    ExpirationTime = 600000

    def __init__(self, port=None, devices=None, loop=SharedLoop):
        super(VirtualDeviceAdapter, self).__init__(name=__name__, loop=loop)

        loaded_devs = {}

        if devices is None:
            devices = []

        if port is not None:
            devs = port.split(';')

            for dev in devs:
                name, _sep, config = dev.partition('@')

                if len(config) == 0:
                    config = None

                loaded_dev = self._load_device(name, config)

                if not self._validate_device(loaded_dev):
                    raise ArgumentError("Device type cannot be loaded on this adapter", name=name)

                loaded_devs[loaded_dev.iotile_id] = loaded_dev

        # Allow explicitly passing created VirtualDevice subclasses
        for dev in devices:
            if not self._validate_device(dev):
                raise ArgumentError("Device type cannot be loaded on this adapter", name=name)

            loaded_devs[dev.iotile_id] = dev

        self.devices = loaded_devs
        self.scan_interval = self.get_config('scan_interval', 1.0)
        self.last_scan = None
        self._logger = logging.getLogger(__name__)

        self.set_config('probe_required', True)
        self.set_config('probe_supported', True)

    async def start(self):
        for dev in self.devices.values():
            await self._loop.run_in_executor(dev.start, VirtualAdapterAsyncChannel(self, dev.iotile_id))

    # pylint:disable=unused-argument;The name needs to remain without an _ for subclasses to override
    @classmethod
    def _validate_device(cls, device):
        """Hook for subclases to ensure that only specific kinds of devices are loaded.

        The default implementation just returns True, allowing all virtual devices.

        Returns:
            bool: Whether the virtual device is allowed to load.
        """

        return True

    def _load_device(self, name, config):
        """Load a device either from a script or from an installed module"""

        if config is None:
            config_dict = {}
        elif isinstance(config, dict):
            config_dict = config
        elif config[0] == '#':
            # Allow passing base64 encoded json directly in the port string to ease testing.
            import base64
            config_str = str(base64.b64decode(config[1:]), 'utf-8')
            config_dict = json.loads(config_str)
        else:
            try:
                with open(config, "r") as conf:
                    data = json.load(conf)
            except IOError as exc:
                raise ArgumentError("Could not open config file", error=str(exc), path=config)

            if 'device' not in data:
                raise ArgumentError("Invalid configuration file passed to VirtualDeviceAdapter",
                                    device_name=name, config_path=config, missing_key='device')

            config_dict = data['device']

        reg = ComponentRegistry()

        if name.endswith('.py'):
            _name, device_factory = reg.load_extension(name, class_filter=BaseVirtualDevice, unique=True)
            return _instantiate_virtual_device(device_factory, config_dict, self._loop)

        seen_names = []
        for device_name, device_factory in reg.load_extensions('iotile.virtual_device',
                                                               class_filter=BaseVirtualDevice,
                                                               product_name="virtual_device"):
            if device_name == name:
                return _instantiate_virtual_device(device_factory, config_dict, self._loop)

            seen_names.append(device_name)

        raise ArgumentError("Could not find virtual_device by name", name=name, known_names=seen_names)

    def can_connect(self):
        """Return whether this device adapter can accept another connection."""

        return True

    async def connect(self, conn_id, connection_string):
        """Asynchronously connect to a device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            connection_string (string): A DeviceAdapter specific string that can be used to connect to
                a device using this DeviceAdapter.
            callback (callable): A function that will be called when the connection attempt finishes as
                callback(conection_id, adapter_id, success: bool, failure_reason: string or None)
        """

        id_number = int(connection_string)
        if id_number not in self.devices:
            raise DeviceAdapterError(conn_id, 'connect', 'device not found')

        if self._get_conn_id(connection_string) is not None:
            raise DeviceAdapterError(conn_id, 'connect', 'device already connected')

        dev = self.devices[id_number]

        if dev.connected:
            raise DeviceAdapterError(conn_id, 'connect', 'device already connected')

        await dev.connect()

        self._setup_connection(conn_id, connection_string)
        self._track_property(conn_id, 'device', dev)

    async def disconnect(self, conn_id):
        """Asynchronously disconnect from a connected device

        Args:
            conn_id (int): A unique identifier that will refer to this connection
            callback (callback): A callback that will be called as
                callback(conn_id, adapter_id, success, failure_reason)
        """

        self._ensure_connection(conn_id, True)

        dev = self._get_property(conn_id, 'device')
        await dev.disconnect()

        self._teardown_connection(conn_id)

    async def open_interface(self, conn_id, interface):
        self._ensure_connection(conn_id, True)

        dev = self._get_property(conn_id, 'device')
        conn_string = self._get_property(conn_id, 'connection_string')

        result = await dev.open_interface(interface)

        if interface == 'streaming' and result is not None:
            for report in result:
                await self.notify_event(conn_string, 'report', report)
        elif interface == 'tracing' and result is not None:
            for trace in result:
                await self.notify_event(conn_string, 'trace', trace)
        elif result is not None:
            self._logger.warning("Discarding non-None result from open_interface call: %s", result)

    async def close_interface(self, conn_id, interface):
        self._ensure_connection(conn_id, True)

        dev = self._get_property(conn_id, 'device')
        await dev.close_interface(interface)

    async def send_rpc(self, conn_id, address, rpc_id, payload, timeout):
        """Asynchronously send an RPC to this IOTile device.

        Args:
            conn_id (int): A unique identifier that will refer to this connection
            address (int): the address of the tile that we wish to send the RPC to
            rpc_id (int): the 16-bit id of the RPC we want to call
            payload (bytearray): the payload of the command
            timeout (float): the number of seconds to wait for the RPC to execute
        """

        self._ensure_connection(conn_id, True)
        dev = self._get_property(conn_id, 'device')

        try:
            return await dev.async_rpc(address, rpc_id, bytes(payload))
        except VALID_RPC_EXCEPTIONS:
            raise
        except Exception:
            self._logger.exception("Exception inside rpc %d:0x%04X, payload=%s",
                                   address, rpc_id, payload)
            raise

    async def send_script(self, conn_id, data):
        """Asynchronously send a script to this IOTile device.

        Args:
            conn_id (int): A unique identifier that will refer to this connection
            data (bytes or bytearray): the script to send to the device
        """

        self._ensure_connection(conn_id, True)
        dev = self._get_property(conn_id, 'device')
        conn_string = self._get_property(conn_id, 'connection_string')

        # Simulate some progress callbacks (0, 50%, 100%)
        await self.notify_progress(conn_string, 'script', 0, len(data))
        await self.notify_progress(conn_string, 'script', len(data) // 2, len(data))
        await self.notify_progress(conn_string, 'script', len(data), len(data))

        dev.script = data

    async def debug(self, conn_id, name, cmd_args):
        """Send a debug command to a device.

        This method responds to a single command 'inspect_property' that takes
        the name of a propery on the device and returns its value.  The
        ``cmd_args`` dict should have a single key: 'properties' that is a
        list of strings with the property names that should be returned.

        Those properties are all queried and their result returned.

        The result is a dict that maps property name to value.  There is a
        progress event generated for every property whose purpose is primarily
        to allow for testing the progress system of a device server.

        See :meth:`AbstractDeviceAdapter.debug`.
        """

        if name == 'heartbeat':
            return {'alive': True}

        self._ensure_connection(conn_id, True)
        dev = self._get_property(conn_id, 'device')
        conn_string = self._get_property(conn_id, 'connection_string')

        if name != 'inspect_property':
            raise DeviceAdapterError(conn_id, 'debug', 'operation {} not supported'.format(name))

        properties = cmd_args.get('properties', [])

        result = {}

        for i, prop in enumerate(properties):
            if not hasattr(dev, prop):
                raise DeviceAdapterError(conn_id, 'debug', 'property {} did not exist'.format(prop))

            value = getattr(dev, prop)
            result[prop] = value

            await self.notify_progress(conn_string, 'debug', i + 1, len(properties))

        return result

    async def _send_scan_event(self, device):
        """Send a scan event from a device."""

        conn_string = str(device.iotile_id)
        info = {
            'connection_string': conn_string,
            'uuid': device.iotile_id,
            'signal_strength': 100,
            'validity_period': self.ExpirationTime
        }

        await self.notify_event(conn_string, 'device_seen', info)

    async def stop(self):
        for dev in self.devices.values():
            await self._loop.run_in_executor(dev.stop)

    async def probe(self):
        """Send advertisements for all connected virtual devices."""

        for dev in self.devices.values():
            await self._send_scan_event(dev)


def _instantiate_virtual_device(factory, config, loop):
    """Safely instantiate a virtualdevice passing a BackgroundEventLoop if necessary."""

    kwargs = dict()

    sig = inspect.signature(factory)
    if 'loop' in sig.parameters:
        kwargs['loop'] = loop

    return factory(config, **kwargs)



import struct
import sys
import warnings

warnings.warn("unpack was a legacy shim for python < 2.7.5 and will be removed in a future iotile-core release", DeprecationWarning)

def unpack(fmt, arg):
    """A shim around struct.unpack to allow it to work on python 2.7.3."""

    if isinstance(arg, bytearray) and not (sys.version_info >= (2, 7, 5)):
        return struct.unpack(fmt, str(arg))

    return struct.unpack(fmt, arg)



"""An enhanced reflash controller record."""

import struct
from iotile.core.exceptions import ArgumentError
from ..record import UpdateRecord, MatchQuality


class EnhancedReflashControllerRecord(UpdateRecord):
    """Enhanced Reflash for an IOTile controller.

        This record is based off the ReflashControllerRecord. This "enhanced"
        version contains additional parameters to allow for a more advanced
        way to reflash an IOTile device. The normal ReflashControllerRecord
        is simple and flashes binary data to flash offset. This record has
        options to allow for settings and flags to allow for future expansion.

        Args:
            raw_data (bytearray): The raw binary firmware data that we should
                program.
            base_address (int): The absolute memory offset at which raw_data 
                starts.
            image_type (int): *The type of firmware image
            flags (int): Specific flags to process
            compression_type (int): *How the appended image data is compressed
            compression_settings_length (int): *Length of compression settings
            compression_settings (bytearray): *Settings for compression
            preinstall_checks (bytearray): *Options to check during install 

        Note:
            The arguments marked with an asterisk(*) means that those features
            have not been implemented yet. This is just allowing for future
            expansion.
    """

    RecordType = 6
    RecordHeaderLength = 96

    def __init__(self, raw_data, base_address, image_type=0, flags=0,
                 compression_type=0, compression_settings_length=0,
                 compression_settings=bytearray(16),
                 preinstall_checks=bytearray(64)):
        self.raw_data = raw_data
        self.base_address = base_address
        self.image_type = image_type
        self.flags = flags
        self.compression_type = compression_type
        self.compression_settings_length = compression_settings_length
        self.compression_settings = compression_settings
        self.preinstall_checks = preinstall_checks

    def encode_contents(self):
        """Encode the contents of the enhanced reflash record.

        Returns:
            bytearray: The encoded contents
        """

        header = struct.pack("<LLLBBBB16s64s", self.base_address,
                             len(self.raw_data), len(self.raw_data),
                             self.image_type, self.flags, self.compression_type,
                             self.compression_settings_length,
                             self.compression_settings, self.preinstall_checks)

        return bytearray(header) + self.raw_data

    @classmethod
    def MatchType(cls):
        """Return the record type that this record matches.

        All records must match an 8-bit record type field that is used to
        decode a binary script.  Note that multiple records may match the same
        8-bit record type if they have different levels of specificity.

        Returns:
            int: The single record type that this record matches.
        """

        return EnhancedReflashControllerRecord.RecordType

    @classmethod
    def MatchQuality(cls, record_data, record_count=1):
        """Check how well this record matches the given binary data.

        This function will only be called if the record matches the type code
        given by calling MatchType() and this function should check how well
        this record matches and return a quality score between 0 and 100, with
        higher quality matches having higher scores.  The default value should
        be MatchQuality.GenericMatch which is 50.  If this record does not
        match at all, it should return MatchQuality.NoMatch.

        Many times, only a single record type will match a given binary record
        but there are times when multiple different logical records produce
        the same type of record in a script, such as set_version and
        set_userkey both producing a call_rpc record with different RPC
        values.  The MatchQuality method is used to allow for rich decoding
        of such scripts back to the best possible record that created them.

        Args:
            record_data (bytearay): The raw record that we should check for
                a match.
            record_count (int): The number of binary records that are included
                in record_data.

        Returns:
            int: The match quality between 0 and 100.  You should use the
                constants defined in MatchQuality as much as possible.
        """

        if record_count > 1:
            return MatchQuality.NoMatch

        return MatchQuality.GenericMatch

    @classmethod
    def FromBinary(cls, record_data, record_count=1):
        """Create an UpdateRecord subclass from binary record data.

        This should be called with a binary record blob (NOT including the
        record type header) and it will decode it into a ReflashControllerRecord.

        Args:
            record_data (bytearray): The raw record data that we wish to parse
                into an UpdateRecord subclass NOT including its 8 byte record header.
            record_count (int): The number of records included in record_data.

        Raises:
            ArgumentError: If the record_data is malformed and cannot be parsed.

        Returns:
            ReflashControllerRecord: The decoded reflash tile record.
        """

        if len(record_data) < EnhancedReflashControllerRecord.RecordHeaderLength:
            raise ArgumentError("Record was too short to contain a full enhanced reflash record header",
                                length=len(record_data), header_length=EnhancedReflashControllerRecord.RecordHeaderLength) 

        base_address, encoded_length, total_length, image_type, flags,\
        compression_type, compression_settings_length, compression_settings,\
        preinstall_checks  = struct.unpack_from("<LLLBBBB16s64s", record_data)

        bindata = record_data[EnhancedReflashControllerRecord.RecordHeaderLength:]
        if len(bindata) != total_length:
            raise ArgumentError("Embedded firmware length did not agree with actual length of embeded data",
                                length=len(bindata), embedded_length=total_length)

        return EnhancedReflashControllerRecord(bindata, base_address,
                                               image_type=image_type,
                                               flags=flags,
                                               compression_type=compression_type,
                                               compression_settings_length=compression_settings_length,
                                               compression_settings=compression_settings,
                                               preinstall_checks=preinstall_checks)

    def __eq__(self, other):
        if not isinstance(other, EnhancedReflashControllerRecord):
            return False

        return self.raw_data == other.raw_data and\
               self.base_address == other.base_address and\
               self.image_type == other.image_type and\
               self.flags == other.flags and\
               self.compression_type == other.compression_type and\
               self.compression_settings_length == other.compression_settings_length and\
               self.compression_settings == other.compression_settings and\
               self.preinstall_checks == other.preinstall_checks


    def __str__(self):
        return "Enhanced Reflash controller with %d (0x%X) bytes starting at base_address %d (0x%X)" %\
            (len(self.raw_data), len(self.raw_data), self.base_address, self.base_address)



"""Standard implementation of AbstractDeviceAdapter."""

import logging
from iotile.core.exceptions import ArgumentError
from iotile.core.utilities import SharedLoop

from .abstract import AbstractDeviceAdapter
from .mixin_conndata import PerConnectionDataMixin
from .mixin_notifications import BasicNotificationMixin
from ...exceptions import DeviceAdapterError

_MISSING = object()


class StandardDeviceAdapter(PerConnectionDataMixin,
                            BasicNotificationMixin,
                            AbstractDeviceAdapter):
    """Standard basic implementation of AbstractDeviceAdapter.

    This class adds utility and convenience functions that are typically
    needed for building real DeviceAdapters as well as base implementations of
    all required :class:AbstractDeviceAdapter methods.  The base
    implementations don't do anything except raise an error saying the
    operation is not supported so subclasses can add features one at at a time
    until they support all of the chosen interfaces.

    Most real DeviceAdapters should inherit from this class rather than
    directly from AbstractDeviceAdapter unless you are doing some really
    special and know why it's better to avoid the helper methods inside this
    class.


    Key Features
    ------------

    1. A default implementation of ``get_config`` and ``set_config`` based on
       an internal dictionary.  This is typically all you need for a normal
       DeviceAdapter.

    2. A per-connection dictionary that you can use to store and fetch information
       that is needed on a per connection basis.  See :meth:`_get_property` and
       :meth:`_track_property`.  You can also use :meth:`_ensure_connection` as
       a simple way to fail a method if it requires a valid connection.  This check
       happens automatically inside :meth:`_get_property`.  This is provided by the
       mixin class :class:`PerConnectionDataMixin`.

    3. A default implementation of notification callbacks using a BackgroundEventLoop
       passed into the constructor.  This is provided by the mixin class
       :class:`BasicNotificationMixin`.

    If you want to include one or more of those mixins in a class that does
    not inherit from StandardDeviceAdapter, you are free to use them directly.


    Subclassing
    -----------

    The only two methods that you need to provide to subclass StandardDeviceAdapter are:

    - ``async def start()``: Start your device adapter
    - ``async def stop()``: Stop your device adapter

    All other required methods have safe default implementations.  When you
    want to send notifications on events, you should await
    :meth:`_notify_event`.


    More Information
    ----------------

    Most of the methods you need to implement and the contracts that they
    must satisfy is listed with :class:`abstract.AbstractDeviceAdapter`.

    - See :class:`abstract.AbstractDeviceAdapter`
    - See :mod:`iotile.core.hw.adapter.abstract`


    Args:
        loop (BackgroundEventLoop): The event loop that we should use to
            run our notification callbacks.
        name (str): Optional name for the logger
    """

    MAX_CONNECTIONS = 1
    """The default number of maximum simultaneous connections we support."""

    def __init__(self, name=__name__, loop=SharedLoop):
        PerConnectionDataMixin.__init__(self)
        BasicNotificationMixin.__init__(self, loop)
        AbstractDeviceAdapter.__init__(self)

        self._logger = logging.getLogger(name)
        self._next_conn_id = 0
        self._config = {}

    def get_config(self, name, default=_MISSING):
        """Get a configuration setting from this DeviceAdapter.

        See :meth:`AbstractDeviceAdapter.get_config`.
        """

        val = self._config.get(name, default)
        if val is _MISSING:
            raise ArgumentError("DeviceAdapter config {} did not exist and no default".format(name))

        return val

    def set_config(self, name, value):
        """Adjust a configuration setting on this DeviceAdapter.

        See :meth:`AbstractDeviceAdapter.set_config`.
        """

        self._config[name] = value

    def can_connect(self):
        """Return whether this device adapter can accept another connection.

        We check the config property 'max_connections' with the default of
        self.MAX_CONNECTIONS and return True if there are fewer active
        connections than that maximum number.

        See :meth:`AbstractDeviceAdapter.can_connect`.
        """

        return len(self._connections) < self.get_config('max_connections', self.MAX_CONNECTIONS)

    def unique_conn_id(self):
        """Return a new, unique connection id.

        See :meth:`AbstractDeviceAdapter.unique_conn_id`.
        """

        conn_id = self._next_conn_id
        self._next_conn_id += 1

        return conn_id

    async def start(self):
        """Start the device adapter.

        See :meth:`AbstractDeviceAdapter.start`.
        """

    async def stop(self):
        """Stop the device adapter.

        See :meth:`AbstractDeviceAdapter.stop`.
        """

    async def connect(self, conn_id, connection_string):
        """Connect to a device.

        See :meth:`AbstractDeviceAdapter.connect`.
        """

        raise DeviceAdapterError(conn_id, 'connect', 'not supported')

    async def disconnect(self, conn_id):
        """Disconnect from a connected device.

        See :meth:`AbstractDeviceAdapter.disconnect`.
        """

        raise DeviceAdapterError(conn_id, 'disconnect', 'not supported')

    async def open_interface(self, conn_id, interface):
        """Open an interface on an IOTile device.

        See :meth:`AbstractDeviceAdapter.open_interface`.
        """

        raise DeviceAdapterError(conn_id, 'open_interface {}'.format(interface),
                                 'not supported')

    async def close_interface(self, conn_id, interface):
        """Close an interface on this IOTile device.

        See :meth:`AbstractDeviceAdapter.close_interface`.
        """

        raise DeviceAdapterError(conn_id, 'close_interface {}'.format(interface),
                                 'not supported')

    async def probe(self):
        """Probe for devices connected to this adapter.

        See :meth:`AbstractDeviceAdapter.probe`.
        """

        raise DeviceAdapterError(None, 'probe', 'not supported')

    async def send_rpc(self, conn_id, address, rpc_id, payload, timeout):
        """Send an RPC to a device.

        See :meth:`AbstractDeviceAdapter.send_rpc`.
        """

        raise DeviceAdapterError(conn_id, 'send rpc', 'not supported')

    async def debug(self, conn_id, name, cmd_args):
        """Send a debug command to a device.

        See :meth:`AbstractDeviceAdapter.debug`.
        """

        raise DeviceAdapterError(conn_id, 'debug', 'not supported')

    async def send_script(self, conn_id, data):
        """Send a a script to a device.

        See :meth:`AbstractDeviceAdapter.send_script`.
        """

        raise DeviceAdapterError(conn_id, 'send_script', 'not supported')



"""Base class for all virtual tiles."""

from iotile.core.exceptions import ArgumentError
from iotile.core.dev import ComponentRegistry
from .common_types import RPCDispatcher


class BaseVirtualTile(RPCDispatcher):
    """A virtual tile.

    Tiles have their own RPC based API and can run background tasks that
    stream or trace data asynchronously to any user-called RPC by using the
    ``channel`` member.

    Args:
        address (int): The address of this tile in the VirtualIOTIleDevice
            that contains it
    """

    def __init__(self, address):
        super(BaseVirtualTile, self).__init__()

        self.address = address
        self.channel = None

    def start(self, channel=None):
        """Start any activity on this tile.

        Subclasses that have background workers or other activity
        can override this method to start their background work.

        Callers should pass an :class:`VirtualAdapterAsyncChannel` object as
        the channel parameter to allow tiles to stream and trace data.

        Args:
            channel (VirtualAdapterAsyncChannel): An optional channel for
                this tile to send stream and trace data asynchronously
                out of the simulated device.
        """

        self.channel = channel

    def stop(self):
        """Stop any background work on this tile."""

    @classmethod
    def FindByName(cls, name):
        """Find an installed VirtualTile by name.

        This function searches for installed virtual tiles
        using the pkg_resources entry_point `iotile.virtual_tile`.

        If name is a path ending in .py, it is assumed to point to
        a module on disk and loaded directly rather than using
        pkg_resources.

        Args:
            name (str): The name of the tile to search
                for.

        Returns:
            VirtualTile class: A virtual tile subclass that can be
                instantiated to create a virtual tile.
        """

        if name.endswith('.py'):
            return cls.LoadFromFile(name)

        reg = ComponentRegistry()
        for _name, tile in reg.load_extensions('iotile.virtual_tile', name_filter=name, class_filter=BaseVirtualTile):
            return tile

        raise ArgumentError("VirtualTile could not be found by name", name=name)

    @classmethod
    def LoadFromFile(cls, script_path):
        """Import a virtual tile from a file rather than an installed module

        script_path must point to a python file ending in .py that contains exactly one
        VirtualTile class definition.  That class is loaded and executed as if it
        were installed.

        To facilitate development, if there is a proxy object defined in the same
        file, it is also added to the HardwareManager proxy registry so that it
        can be found and used with the device.

        Args:
            script_path (string): The path to the script to load

        Returns:
            VirtualTile: A subclass of VirtualTile that was loaded from script_path
        """

        _name, dev = ComponentRegistry().load_extension(script_path, class_filter=BaseVirtualTile, unique=True)
        return dev



"""IOTileReport subclass for readings packaged as individual readings"""

import datetime
import struct
import hashlib
import hmac
from enum import IntEnum
from .report import IOTileReport, IOTileReading
from iotile.core.utilities.packed import unpack
from iotile.core.exceptions import NotFoundError, ExternalError
from iotile.core.hw.auth.auth_provider import AuthProvider
from iotile.core.hw.auth.auth_chain import ChainedAuthProvider

class ReportSignatureFlags(IntEnum):
    """Types of stream report signature"""
    SIGNED_WITH_HASH = 0
    SIGNED_WITH_USER_KEY = 1
    SIGNED_WITH_DEVICE_KEY = 2

class SignedListReport(IOTileReport):
    """A report that consists of a signed list of readings.

    Args:
        rawreport (bytearray): The raw data of this report
    """

    ReportType = 1

    def __init__(self, rawreport, **kwargs):
        super(SignedListReport, self).__init__(rawreport, signed=True, encrypted=False, **kwargs)

    @classmethod
    def KeyTypeToStreamType(cls, key_type):
        """Converts key type into the type of signed report that can be encrypted with this key

        Args:
            int: a type of the key, see AuthProvider

        Returns:
            ReportSignatureFlags: a type of the stream report
        """
        if key_type == AuthProvider.NoKey:
            return ReportSignatureFlags.SIGNED_WITH_HASH
        elif key_type == AuthProvider.UserKey:
            return ReportSignatureFlags.SIGNED_WITH_USER_KEY
        elif key_type == AuthProvider.DeviceKey:
            return ReportSignatureFlags.SIGNED_WITH_DEVICE_KEY
        else:
            raise ExternalError("Unsupported key type {}".format(key_type))

    @classmethod
    def StreamTypeToKeyType(cls, stream_type):
        """Converts the stream type into the key type that was used to encryped this stream report

        Args:
            ReportSignatureFlags: a type of the stream report

        Returns:
            int: type of key, see AuthProvider
        """

        if stream_type == ReportSignatureFlags.SIGNED_WITH_HASH:
            return AuthProvider.NoKey
        elif stream_type == ReportSignatureFlags.SIGNED_WITH_USER_KEY:
            return AuthProvider.UserKey
        elif stream_type == ReportSignatureFlags.SIGNED_WITH_DEVICE_KEY:
            return AuthProvider.DeviceKey
        else:
            raise ExternalError("Unsupported stream signature type {}".format(stream_type))

    @classmethod
    def HeaderLength(cls):
        """Return the length of a header needed to calculate this report's length

        Returns:
            int: the length of the needed report header
        """

        return 20

    @classmethod
    def ReportLength(cls, header):
        """Given a header of HeaderLength bytes, calculate the size of this report"""

        first_word, = unpack("<L", header[:4])

        length = (first_word >> 8)
        return length

    @classmethod
    def VerifyReport(cls, device_id, auth_provider, key_type, data, signature, **kwargs):
        """Verify a buffer of report data on behalf of a device.

        Args:
            device_id (int): The id of the device that we should encrypt for
            root (int): The root key type that should be used to generate the report
            data (bytearray): The data that we should verify
            signature (bytearray): The signature attached to data that we should verify
            **kwargs: There are additional specific keyword args that are required
                depending on the root key used.  Typically, you must specify
                - report_id (int): The report id
                - sent_timestamp (int): The sent timestamp of the report

                These two bits of information are used to construct the per report
                signing and encryption key from the specific root key type.

        Returns:
            dict: The result of the verification process must always be a bool under the
                'verified' key, however additional keys may be present depending on the
                signature method used.

        Raises:
            NotFoundError: If the auth provider is not able to verify the data due to
                an error.  If the data is simply not valid, then the function returns
                normally.
        """

        if key_type == AuthProvider.NoKey:
            result = bytearray(hashlib.sha256(data).digest())
        else:
            report_key = auth_provider.get_serialized_key(key_type, device_id, **kwargs)

            message_hash = hashlib.sha256(data).digest()
            hmac_calc = hmac.new(report_key, message_hash, hashlib.sha256)
            result = bytearray(hmac_calc.digest())

        if len(signature) == 0:
            verified = False
        elif len(signature) > len(result):
            verified = False
        elif len(signature) < len(result):
            trunc_result = result[:len(signature)]
            verified = hmac.compare_digest(signature, trunc_result)
        else:
            verified = hmac.compare_digest(signature, result)

        return {'verified': verified, 'bit_length': 8*len(signature)}

    @classmethod
    def DecryptReport(cls, device_id, auth_provider, key_type, data, **kwargs):
        """Decrypt a buffer of report data on behalf of a device.

        Args:
            device_id (int): The id of the device that we should encrypt for
            root (int): The root key type that should be used to generate the report
            data (bytearray): The data that we should decrypt
            **kwargs: There are additional specific keyword args that are required
                depending on the root key used.  Typically, you must specify
                - report_id (int): The report id
                - sent_timestamp (int): The sent timestamp of the report

                These two bits of information are used to construct the per report
                signing and encryption key from the specific root key type.

        Returns:
            dict: The decrypted data and any associated metadata about the data.
                The data itself must always be a bytearray stored under the 'data'
                key, however additional keys may be present depending on the encryption method
                used.

        Raises:
            NotFoundError: If the auth provider is not able to decrypt the data.
        """

        report_key = auth_provider.get_serialized_key(key_type, device_id, **kwargs)

        try:
            from Crypto.Cipher import AES
            import Crypto.Util.Counter
        except ImportError:
            raise NotFoundError

        ctr = Crypto.Util.Counter.new(128)

        # We use AES-128 for encryption
        encryptor = AES.new(bytes(report_key[:16]), AES.MODE_CTR, counter=ctr)

        decrypted = encryptor.decrypt(bytes(data))
        return {'data': decrypted}

    @classmethod
    def EncryptReport(cls, device_id, auth_provider, key_type, data, **kwargs):
        """Encrypt a buffer of report data on behalf of a device.

        Args:
            device_id (int): The id of the device that we should encrypt for
            root (int): The root key type that should be used to generate the report
            data (bytearray): The data that we should decrypt
            **kwargs: There are additional specific keyword args that are required
                depending on the root key used.  Typically, you must specify
                - report_id (int): The report id
                - sent_timestamp (int): The sent timestamp of the report

                These two bits of information are used to construct the per report
                signing and encryption key from the specific root key type.

        Returns:
            dict: The encrypted data and any associated metadata about the data.
                The data itself must always be a bytearray stored under the 'data'
                key, however additional keys may be present depending on the encryption method
                used.

        Raises:
            NotFoundError: If the auth provider is not able to decrypt the data.
        """

        report_key = auth_provider.get_serialized_key(key_type, device_id, **kwargs)

        try:
            from Crypto.Cipher import AES
            import Crypto.Util.Counter
        except ImportError:
            raise NotFoundError

        # We use AES-128 for encryption
        ctr = Crypto.Util.Counter.new(128)
        encryptor = AES.new(bytes(report_key[:16]), AES.MODE_CTR, counter=ctr)

        encrypted = encryptor.encrypt(bytes(data))
        return {'data': encrypted}

    @classmethod
    def SignReport(cls, device_id, auth_provider, key_type, data, **kwargs):
        """Sign a buffer of report data on behalf of a device.

        Args:
            device_id (int): The id of the device that we should encrypt for
            root (int): The root key type that should be used to generate the report
            data (bytearray): The data that we should sign
            **kwargs: There are additional specific keyword args that are required
                depending on the root key used.  Typically, you must specify
                - report_id (int): The report id
                - sent_timestamp (int): The sent timestamp of the report

                These two bits of information are used to construct the per report
                signing and encryption key from the specific root key type.

        Returns:
            dict: The signature and any associated metadata about the signature.
                The signature itself must always be a bytearray stored under the
                'signature' key, however additional keys may be present depending
                on the signature method used.

        Raises:
            NotFoundError: If the auth provider is not able to sign the data.
        """

        if key_type == AuthProvider.NoKey:
            result = bytearray(hashlib.sha256(data).digest())
        else:
            report_key = auth_provider.get_serialized_key(key_type, device_id, **kwargs)

            # We sign the SHA256 hash of the message
            message_hash = hashlib.sha256(data).digest()
            hmac_calc = hmac.new(report_key, message_hash, hashlib.sha256)
            result = bytearray(hmac_calc.digest())

        return {'signature': result, 'root_key': key_type}


    @classmethod
    def FromReadings(cls, uuid, readings, root_key=AuthProvider.NoKey, signer=None,
                     report_id=IOTileReading.InvalidReadingID, selector=0xFFFF, streamer=0, sent_timestamp=0):
        """Generate an instance of the report format from a list of readings and a uuid.

        The signed list report is created using the passed readings and signed using the specified method
        and AuthProvider.  If no auth provider is specified, the report is signed using the default authorization
        chain.

        Args:
            uuid (int): The uuid of the deviec that this report came from
            readings (list): A list of IOTileReading objects containing the data in the report
            root_key (int): The key that should be used to sign the report (must be supported
                by an auth_provider)
            signer (AuthProvider): An optional preconfigured AuthProvider that should be used to sign this
                report.  If no AuthProvider is provided, the default ChainedAuthProvider is used.
            report_id (int): The id of the report.  If not provided it defaults to IOTileReading.InvalidReadingID.
                Note that you can specify anything you want for the report id but for actual IOTile devices
                the report id will always be greater than the id of all of the readings contained in the report
                since devices generate ids sequentially.
            selector (int): The streamer selector of this report.  This can be anything but if the report came from
                a device, it would correspond with the query the device used to pick readings to go into the report.
            streamer (int): The streamer id that this reading was sent from.
            sent_timestamp (int): The device's uptime that sent this report.
        """

        lowest_id = IOTileReading.InvalidReadingID
        highest_id = IOTileReading.InvalidReadingID

        report_len = 20 + 16*len(readings) + 24
        len_low = report_len & 0xFF
        len_high = report_len >> 8

        unique_readings = [x.reading_id for x in readings if x.reading_id != IOTileReading.InvalidReadingID]
        if len(unique_readings) > 0:
            lowest_id = min(unique_readings)
            highest_id = max(unique_readings)

        signature_flags = SignedListReport.KeyTypeToStreamType(root_key)
        header = struct.pack("<BBHLLLBBH", cls.ReportType, len_low, len_high, uuid, report_id,
                             sent_timestamp, signature_flags, streamer, selector)
        header = bytearray(header)

        packed_readings = bytearray()

        for reading in readings:
            packed_reading = struct.pack("<HHLLL", reading.stream, 0, reading.reading_id,
                                         reading.raw_time, reading.value)
            packed_readings += bytearray(packed_reading)

        footer_stats = struct.pack("<LL", lowest_id, highest_id)

        if signer is None:
            signer = ChainedAuthProvider()

        # If we are supposed to encrypt this report, do the encryption
        if root_key != signer.NoKey:
            enc_data = packed_readings

            try:
                result = SignedListReport.EncryptReport(uuid, signer, root_key, enc_data, report_id=report_id,
                                               sent_timestamp=sent_timestamp)
            except NotFoundError:
                raise ExternalError("Could not encrypt report because no AuthProvider supported "
                                    "the requested encryption method for the requested device",
                                    device_id=uuid, root_key=root_key)

            signed_data = header + result['data'] + footer_stats
        else:
            signed_data = header + packed_readings + footer_stats

        try:
            signature = SignedListReport.SignReport(uuid, signer, root_key, signed_data, report_id=report_id,
                                           sent_timestamp=sent_timestamp)
        except NotFoundError:
            raise ExternalError("Could not sign report because no AuthProvider supported the requested "
                                "signature method for the requested device", device_id=uuid, root_key=root_key)

        footer = struct.pack("16s", bytes(signature['signature'][:16]))
        footer = bytearray(footer)

        data = signed_data + footer
        return SignedListReport(data)

    def decode(self):
        """Decode this report into a list of readings
        """

        fmt, len_low, len_high, device_id, report_id, sent_timestamp, signature_flags, \
        origin_streamer, streamer_selector = unpack("<BBHLLLBBH", self.raw_report[:20])

        assert fmt == 1
        length = (len_high << 8) | len_low

        self.origin = device_id
        self.report_id = report_id
        self.sent_timestamp = sent_timestamp
        self.origin_streamer = origin_streamer
        self.streamer_selector = streamer_selector
        self.signature_flags = signature_flags

        assert len(self.raw_report) == length

        remaining = self.raw_report[20:]
        assert len(remaining) >= 24
        readings = remaining[:-24]
        footer = remaining[-24:]

        lowest_id, highest_id, signature = unpack("<LL16s", footer)
        signature = bytearray(signature)

        self.lowest_id = lowest_id
        self.highest_id = highest_id
        self.signature = signature

        signed_data = self.raw_report[:-16]
        signer = ChainedAuthProvider()

        key_type = self.StreamTypeToKeyType(signature_flags)
        self.encrypted = (key_type != AuthProvider.NoKey)

        try:
            verification = SignedListReport.VerifyReport(device_id, signer, key_type, signed_data, signature,
                                                report_id=report_id, sent_timestamp=sent_timestamp)
            self.verified = verification['verified']
        except NotFoundError:
            self.verified = False

        # If we were not able to verify the report, do not try to parse or decrypt it since we
        # can't guarantee who it came from.
        if not self.verified:
            return [], []

        # If the report is encrypted, try to decrypt it before parsing the readings
        if self.encrypted:
            try:
                result = SignedListReport.DecryptReport(device_id, signer, key_type, readings,
                                               report_id=report_id, sent_timestamp=sent_timestamp)
                readings = result['data']
            except NotFoundError:
                return [], []

        # Now parse all of the readings
        # Make sure this report has an integer number of readings
        assert (len(readings) % 16) == 0

        time_base = self.received_time - datetime.timedelta(seconds=sent_timestamp)
        parsed_readings = []

        for i in range(0, len(readings), 16):
            reading = readings[i:i+16]
            stream, _, reading_id, timestamp, value = unpack("<HHLLL", reading)

            parsed = IOTileReading(timestamp, stream, value, time_base=time_base, reading_id=reading_id)
            parsed_readings.append(parsed)

        return parsed_readings, []



""" Interface that provides access to keys"""

import hashlib
import struct
import hmac
from iotile.core.exceptions import NotFoundError
from .auth_provider import AuthProvider

class RootKeyAuthProvider(AuthProvider):
    """Base class for Providers that derive temporary keys from root

        Derived classes should implement get_root_key function
    """

    def get_root_key(self, key_type, device_id):
        """Get the actual root key of a given type for a device.

        Args:
            key_type (int): no key, user key or device key
            device_id (int): UUID of the device

        Returns:
            bytearray: the root key
        """
        raise NotImplementedError()

    def get_rotated_key(self, key_type, device_id, **rotation_info):
        """Get a key that is only valid for a limit period of time.

        Args:
            key_type (int): no key, user key or device key
            device_id (int): UUID of the device
            key_info (dict): data that describes the conditions when
                the key is rotated. For example, for a broadcast report key
                it may be the reboot counter of the
                device, the current uptime and the rotation interval of the key.

        Returns:
            bytearray: the rotated key
        """
        counter = rotation_info.get("reboot_counter", None)
        interval_power = rotation_info.get("rotation_interval_power", None)
        timestamp = rotation_info.get("current_timestamp", None)

        if not counter:
            raise NotFoundError('reboot_counter is not provided in EnvAuthProvider get_rotated_key')

        if not interval_power:
            raise NotFoundError('rotation_interval_power is not provided in EnvAuthProvider get_rotated_key')

        if not timestamp:
            raise NotFoundError('timestamp is not provided in EnvAuthProvider get_rotated_key')

        self.verify_key(key_type)

        root_key = self.get_root_key(key_type, device_id)
        reboot_key = self.DeriveRebootKey(root_key, 0, counter)

        temp_key = self.DeriveRotatedKey(reboot_key, timestamp, interval_power)
        return temp_key


    def get_serialized_key(self, key_type, device_id, **key_info):
        """Get a serialized key such for signing a streamer report.

        These keys are designed to only be used once and only provide
            access to the object of key_type with the given serial_number

        Args:
            key_type (int): no key, user key or device key
            device_id (int): UUID of the device
            key_info (dict): data required for key generation.
                It may be report_id and sent_timestamp

        Returns:
            bytearray: the key
        """
        report_id = key_info.get('report_id', None)
        sent_timestamp = key_info.get('sent_timestamp', None)

        if report_id is None or sent_timestamp is None:
            raise NotFoundError('report_id or sent_timestamp is not provided')

        self.verify_key(key_type)

        root_key = self.get_root_key(key_type, device_id)
        report_key = self.DeriveReportKey(root_key, report_id, sent_timestamp)

        return report_key



from .legacy import DeviceAdapter
from .abstract import AbstractDeviceAdapter
from .standard import StandardDeviceAdapter
from .mixin_conndata import PerConnectionDataMixin
from .mixin_notifications import BasicNotificationMixin
from .async_wrapper import AsynchronousModernWrapper

__all__ = ['DeviceAdapter', 'AbstractDeviceAdapter', 'StandardDeviceAdapter',
           'PerConnectionDataMixin', 'BasicNotificationMixin',
           'AsynchronousModernWrapper']



"""An ordered list of authentication providers that are checked in turn to attempt to provide a key"""
from iotile.core.exceptions import NotFoundError, ExternalError
from iotile.core.dev import ComponentRegistry
from .auth_provider import AuthProvider


class ChainedAuthProvider(AuthProvider):
    """An AuthProvider that delegates operations to a list of subproviders.

    Each subprovider is checked in turn and the first one able to satisfy a request
    is used.  If no provider is available that can perform an operation, NotFoundError
    is raised.

    By default, the python entry_point group 'iotile.default_auth_providers' is used
    to build the list of auth providers in the chain.  The entries in that group should
    be tuples of (priority, auth_provider_class, arg_dict) where priority is an integer,
    auth_provider_class is an AuthProvider subclass and arg_dict is a dictionary of
    arguments passed to the constructor of auth_provider.
    """

    def __init__(self, args=None):
        super(ChainedAuthProvider, self).__init__(args)

        # FIXME: Allow overwriting default providers via args
        self._load_installed_providers()

        reg = ComponentRegistry()

        sub_providers = []
        for _, (priority, provider, provider_args) in reg.load_extensions('iotile.default_auth_providers'):
            if provider not in self._auth_factories:
                raise ExternalError("Default authentication provider list references unknown auth provider",
                                    provider_name=provider, known_providers=self._auth_factories.keys())
            configured = self._auth_factories[provider](provider_args)
            sub_providers.append((priority, configured))

        sub_providers.sort(key=lambda x: x[0])
        self.providers = sub_providers

    def _load_installed_providers(self):
        self._auth_factories = {}
        reg = ComponentRegistry()

        for name, entry in reg.load_extensions('iotile.auth_provider'):
            self._auth_factories[name] = entry

    def get_root_key(self, key_type, device_id):
        """Deligates call to auth providers in the chain

        Args:
            key_type (int): see KnownKeyRoots
            device_id (int): uuid of the device

        Returns:
            bytes: the root key
        """
        for _priority, provider in self.providers:
            try:
                return provider.get_root_key(key_type, device_id)
            except NotFoundError:
                pass

        raise NotFoundError("get_serialized_key method is not implemented in any sub_providers")

    def get_serialized_key(self, key_type, device_id, **key_info):
        """Deligates call to auth providers in the chain

        Args:
            key_type (int): see KnownKeyRoots
            device_id (int): uuid of the device
            key_info (dict): specific values for every auth provider

        Returns:
            bytes: the serialized key
        """
        for _priority, provider in self.providers:
            try:
                return provider.get_serialized_key(key_type, device_id, **key_info)
            except NotFoundError:
                pass

        raise NotFoundError("get_serialized_key method is not implemented in any sub_providers")

    def get_rotated_key(self, key_type, device_id, **rotation_info):
        """Deligates call to auth providers in the chain

        Args:
            key_type (int): see KnownKeyRoots
            device_id (int): uuid of the device
            rotation_info (dict): specific value for every auth provider

        Returns:
            bytes: the rotated key
        """
        for _priority, provider in self.providers:
            try:
                return provider.get_rotated_key(key_type, device_id, **rotation_info)
            except NotFoundError:
                pass

        raise NotFoundError("get_rotated_key method is not implemented in any sub_providers")




"""Utilities for simplifying async operations.

This subpackage contains a number of general utility classes for working with
complex asynchronous operations on top of the asyncio package.
"""

from .event_loop import BackgroundEventLoop, SharedLoop, BackgroundTask
from .operation_manager import OperationManager
from .awaitable_dict import AwaitableDict

__all__ = ['BackgroundEventLoop', 'SharedLoop', 'OperationManager',
           'AwaitableDict', 'BackgroundTask']
